<!---+
	Cool function pointer based closure.
--->
<cfcomponent output="false">
	
	<!---+
		Initializes the closure with a function and a context.
		
		ex.
		
		function test( a, b ) { return a + b; }
		
		c = createObject("component","Closure").init(test,variables);
	--->
	<cffunction name="init" access="public" returntype="any" output="false">
		<cfargument name="func" type="any" required="true">
		<cfargument name="context" type="any" required="true">
		
		<cfset variables.func = arguments.func>
		<cfset variables.context = arguments.context>
		<cfset variables.metaData = getMetaData(arguments.func)>
		<cfset variables.paramCount = arrayLen(variables.metaData.parameters)>
	
		<cfset setup()>
						
		<cfreturn this>
	</cffunction>
		
	<!---
		Calls the closure with positional arguments.
		
		ex. clos.call(1,2,3);
	--->
	<cffunction name="call" access="public" returntype="any" output="false">
		
		<cfset var result = "">
		<cfset var args = arguments>
						
		<cfif variables.paramCount eq 1>
			<cfreturn func(args[1],context,metaData)>
		<cfelseif variables.paramCount eq 2>
			<cfreturn func(args[1],args[2],context,metaData)>
		<cfelse>
			<cfreturn evaluate("func(#variables.passedArgs#context,metaData)")>
		</cfif>
			
		<cfif isDefined("result")>
			<cfreturn result>
		</cfif>
	</cffunction>
	
	<!---
		Calls the closure with named arguments.
		
		ex.
		
		args = structNew();
		args.a = 1;
		args.b = 2;
		clos.call(args);
	--->	
	<cffunction name="apply" access="public" returntype="any" output="false">
		<cfargument name="args" type="any" required="true">

		<cfset arguments.variables = variables.context>
		<cfset arguments.metaData = variables.metaData>
				
		<cfreturn func( argumentCollection=args )>
	</cffunction>
	
	<!---
		Computes the argument list for calling the closure faster without as many string ops.
	--->
	<cffunction name="setup" access="private" returntype="any" output="false">
		<cfset var passedArgs = "args[1],args[2],args[3],args[4],args[5],args[6],args[7],">
		<cfset var params = variables.metaData.parameters>
		<cfset var i = "">
		
		<cfif paramCount lte 5>
			<cfset passedArgs = mid(passedArgs,1,8*arrayLen(params))>
		<cfelse>
			<cfloop from="8" to="#paramCount#" index="i">
				<cfset passedArgs &= "args[#i#],">
			</cfloop>
		</cfif>
		
		<cfset variables.passedArgs = passedArgs>
	</cffunction>
	
</cfcomponent>