<cfcomponent output="false">

	<cffunction name="init" access="public" returntype="any" output="false">
		
		<cfset variables.prototype = "">
		<cfset variables.prototypeMethods = structNew()>
		<cfset variables.template = createObject("component",getMetaData(this).name)>
		<cfset variables.prototypeSetupMethods = arrayNew(1)>
		<cfset variables.flattenedVariables = structNew()>
		
		<cfreturn this>
	</cffunction>

	<cffunction name="OnMissingMethod" access="public" returntype="any" output="false">
		<cfargument name="missingMethodName" type="string" required="true">
		<cfargument name="missingMethodArguments" type="any" required="true">
		
		<cfset var local = { result="" }>
		<cfset var method = getMethod(arguments.missingMethodName)>
		
		<cfset this[arguments.missingMethodName] = method>
		<cfset variables.prototypeMethods[arguments.missingMethodName] = method>
				
		<cfinvoke 
			component="#this#" 
			method="#arguments.missingMethodName#" 
			returnvariable="local.result" 
			argumentcollection="#arguments.missingMethodArguments#">
				
		<cfif structKeyExists(local,"result")>
			<cfreturn local.result>
		</cfif>
	</cffunction>
	
	<cffunction name="getMethod" access="public" returntype="any" output="false">
		<cfargument name="name" type="any" required="true">
		
		<cfset var method = "">
		
		<cfif structKeyExists(this,arguments.name) and isCustomFunction(this[arguments.name])>
			<cfreturn this[arguments.name]>
		</cfif>
						
		<cfif hasPrototype()>
			<cfreturn getPrototype().getMethod(arguments.name)>
		</cfif>
		
		<cfthrow type="MissingMethod" message="Method #arguments.name# doesn't exist.">
	</cffunction>
	
	<cffunction name="new" access="public" returntype="any" output="false">
		
		<cfset var setupMethod = "">
		<cfset var object = duplicate(variables.template)>
		
		<cfset structDelete(object,"new")>
		<cfset object.setVariablesScope(variables.flattenedVariables)>
		<cfset object.runInitObjectMethods()>
				
		<cfreturn object>
	</cffunction>
		
	<cffunction name="hasPrototype" access="public" returntype="any" output="false">

		<cfreturn structKeyExists(variables,"prototype") and isObject(variables.prototype)>
	</cffunction>
	
	<cffunction name="setVariablesScope" access="package" returntype="any" output="false">
		<cfargument name="values" type="struct" required="true">
				
		<cfset structAppend(variables,values,false)>
	</cffunction>
	
	<cffunction name="runInitObjectMethods" access="public" returntype="any" output="false">
		<cfset var setupMethod = "">
				
		<cfloop array="#variables.prototypeSetupMethods#" index="setupMethod">
			<cfset setupMethod()>
		</cfloop>
		
	</cffunction>
	
	<cffunction name="flattenVariableScope" access="public" returntype="any" output="false">
		
		<cfset var vars = structNew()>
		<cfset var item = "">
		
		<cfloop collection="#variables#" item="item">
			<cfif not isCustomFunction(variables[item])>
				<cfset vars[item] = variables[item]>
			</cfif>
		</cfloop>
				
		<cfif hasPrototype()>
			<cfset structAppend(vars,getPrototype().flattenVariableScope(),false)>
		</cfif>
		
		<cfreturn vars>
	</cffunction>
	
	<cffunction name="clearPrototype" access="public" returntype="any" output="false">

		<cfset var methodName = "">
		
		<cfloop collection="#variables.prototypeMethods#" item="methodName">
			<cfset structDelete(variables,methodName)>
			<cfset structDelete(this,methodName)>
		</cfloop>
		
		<cfset variables.prototype = "">
	</cffunction>
	
	<cffunction name="findPrototypeSetupMethods" access="public" returntype="array" output="false">
		<cfset var methods = arrayNew(1)>
		
		<cfif hasPrototype()>
			<cfset methods.addAll(getPrototype().findPrototypeSetupMethods())>
		</cfif>
		
		<cfif structKeyExists(variables,"initObject") and isCustomFunction(variables.initObject)>
			<cfset arrayAppend(methods,variables.initObject)>
		</cfif>		
		
		<cfreturn methods>
	</cffunction>
	
	<cffunction name="getPrototype" access="public" returntype="any" output="false">

		<cfreturn variables.prototype>
	</cffunction>
	
	<cffunction name="setPrototype" access="public" returntype="void" output="false">
		<cfargument name="value" type="component" required="true">
		
		<cfset clearPrototype()>
		
		<cfset variables.prototype = arguments.value>
		<cfset variables.prototypeSetupMethods = findPrototypeSetupMethods()>
		<cfset variables.flattenedVariables = flattenVariableScope()>
	</cffunction>

</cfcomponent>